"""
=============
Lookup Tables
=============

Simulations tend to require a large quantity of data to run.  ``vivarium``
provides the :class:`LookupTable` abstraction to make sure the right parameters
are used in the right places.

As discussed in the population :ref:`concept note <population_concept>`,
``vivarium`` uses a :class:`pandas.Index` as the fundamental representation
of the simulated population. The :class:`LookupTable` is then a callable object
that takes in the population index and returns parameters specific to the
individuals represented by that index.
"""
from numbers import Number
from datetime import datetime, timedelta
from typing import Union, List, Tuple, Callable, TypeVar

import pandas as pd

from vivarium.interpolation import Interpolation, ParameterType
from vivarium.framework.population import PopulationView

ScalarValue = TypeVar('ScalarValue', Number, timedelta, datetime)


class InterpolatedTable:
    """A callable that returns the result of an interpolation function over input data.

    Attributes
    ----------
    data :
        The data from which to build the interpolation. Contains the set of
        key_columns, parameter_columns, and value_columns.
    population_view :
        View of the population to be used when the table is called with an index.
    key_columns :
        Column names to be used as categorical parameters in Interpolation
        to select between interpolation functions.
    parameter_columns :
        Column names to be used as continuous parameters in Interpolation.
    value_columns :
        Names of value columns to be interpolated over. All non parameter- and key-
        columns in data.
    interpolation_order :
        Order of interpolation.
    clock :
        Callable for current time in simulation.
    extrapolate:
        Whether or not to extrapolate beyond edges of given bins.

    Notes
    -----
    These should not be created directly. Use the `lookup` method on the builder during setup.
    """
    def __init__(self, data: pd.DataFrame, population_view: PopulationView, key_columns: Union[List[str], Tuple[str]],
                 parameter_columns: ParameterType, value_columns: Union[List[str], Tuple[str]],
                 interpolation_order: int, clock: Callable, extrapolate: bool):

        self.data = data
        self.population_view = population_view
        self.key_columns = key_columns
        self.parameter_columns = parameter_columns
        self.interpolation_order = interpolation_order
        self.value_columns = value_columns
        self.clock = clock
        self.extrapolate = extrapolate
        self.interpolation = Interpolation(data, self.key_columns, self.parameter_columns,
                                           order=self.interpolation_order, extrapolate=self.extrapolate)

    def __call__(self, index: pd.Index) -> pd.DataFrame:
        """Get the interpolated values for the rows in ``index``.

        Parameters
        ----------
        index :
            Index of the population to interpolate for.

        Returns
        -------
        pd.DataFrame
            A table with the interpolated values for the population requested.
        """

        pop = self.population_view.get(index)
        del pop['tracked']
        if 'year' in [col for p in self.parameter_columns for col in p]:
            current_time = self.clock()
            fractional_year = current_time.year
            fractional_year += current_time.timetuple().tm_yday / 365.25
            pop['year'] = fractional_year

        return self.interpolation(pop)

    def __repr__(self):
        return "InterpolatedTable()"


class ScalarTable:
    """A callable that returns a series or dataframe from a single value or list of values with no interpolation.

    Attributes
    ----------
    values :
        The scalar value(s) from which to build table columns.
    value_columns :
        List of string names to be used to name the columns of the table built
        from values.

    Notes
    -----
    These should not be created directly. Use the `lookup` method on the builder during setup.
    """
    def __init__(self, values: Union[List[ScalarValue], Tuple[ScalarValue]], value_columns: Union[List[str], Tuple[str]]):

        self.values = values
        self.value_columns = value_columns

    def __call__(self, index) -> pd.DataFrame:
        """Build a table with ``index`` and columns for each of the scalar values.

        Parameters
        ----------
        index :
            Index of the population to construct table for.

        Returns
        -------
        pd.DataFrame
            A table with a column for each of the scalar values for the population requested.
        """
        if not isinstance(self.values, (list, tuple)):
            values = pd.Series(self.values, index=index, name=self.value_columns[0] if self.value_columns else None)
        else:
            values = dict(zip(self.value_columns, [pd.Series(v, index=index) for v in self.values]))
        return pd.DataFrame(values)

    def __repr__(self):
        return "ScalarTable(value(s)={})".format(self.values)


class LookupTable:
    """Container for ScalarTables/InterpolatedTables over input data.

    Attributes
    ----------
    table : ScalarTable or InterpolatedTable
        callable table created from input data

    Notes
    -----
    These should not be created directly. Use the `lookup` method on the builder during setup.
    """
    def __init__(self, data: Union[ScalarValue, pd.DataFrame, List[ScalarValue], Tuple[ScalarValue]],
                 population_view: Callable, key_columns: Union[List[str], Tuple[str]],
                 parameter_columns: ParameterType, value_columns: Union[List[str], Tuple[str]],
                 interpolation_order: int, clock: Callable, extrapolate: bool):

        validate_parameters(data, key_columns, parameter_columns, value_columns)

        # Note datetime catches pandas timestamps
        if isinstance(data, (Number, datetime, timedelta, list, tuple)):
            self._table = ScalarTable(data, value_columns)
        else:
            callable_parameter_columns = [p[0] for p in parameter_columns]
            view_columns = sorted((set(key_columns) | set(callable_parameter_columns)) - {'year'}) + ['tracked']
            self._table = InterpolatedTable(data, population_view(view_columns), key_columns,
                                            parameter_columns, value_columns, interpolation_order, clock, extrapolate)

    def __call__(self, index) -> Union[pd.DataFrame, pd.Series]:
        """Get the interpolated or scalar table values for the given index.

        Parameters
        ----------
        index : Index
              Index for population view

        Returns
        -------
        pandas.Series if interpolated or scalar values for index are one column,
        pandas.DataFrame if multiple columns
        """
        table_view = self._table(index)
        if len(table_view.columns) == 1:
            return table_view[table_view.columns[0]]
        return table_view

    def __repr__(self):
        return "LookupTable()"


def validate_parameters(data, key_columns, parameter_columns, value_columns):
    if data is None or (isinstance(data, pd.DataFrame) and data.empty) or (isinstance(data, (list, tuple)) and not data):
        raise ValueError("Must supply some data")

    if not isinstance(data, (Number, datetime, timedelta, list, tuple, pd.DataFrame)):
        raise TypeError(f'The only allowable types for data are number, datetime, timedelta, '
                        f'list, tuple, or pandas.DataFrame. You passed {type(data)}.')

    if isinstance(data, (list, tuple)):
        if not value_columns:
            raise ValueError(f'To invoke scalar view with multiple values, you must supply value_columns')
        if len(value_columns) != len(data):
            raise ValueError(f'The number of value columns must match the number of values.'
                             f'You supplied values: {data} and value_columns: {value_columns}')

    if isinstance(data, pd.DataFrame):
        all_parameter_columns = [col for p in parameter_columns for col in p]
        if set(key_columns).intersection(set(all_parameter_columns)):
            raise ValueError(f'There should be no overlap between key columns: {key_columns} '
                             f'and parameter columns: {parameter_columns}.')

        if value_columns:
            data_value_columns = data.columns.difference(set(key_columns) | set(all_parameter_columns))
            if set(value_columns) != set(data_value_columns):
                raise ValueError(f'The value columns you supplied: {value_columns} do not match '
                                 f'the non-parameter columns in the passed data: {data_value_columns}')


class LookupTableManager:
    """Container for LookupTables over input data.

    Notes
    -----
    Client code should never access this class directly. Use ``lookup`` on the builder during setup
    to get references to LookupTable objects.
    """

    configuration_defaults = {
        'interpolation': {
            'order': 0,
            'extrapolate': True
        }
    }

    @property
    def name(self):
        return "lookup_table_manager"

    def setup(self, builder):
        self._pop_view_builder = builder.population.get_view
        self.clock = builder.time.clock()
        self._interpolation_order = builder.configuration.interpolation.order
        self._extrapolate = builder.configuration.interpolation.extrapolate

    def build_table(self, data, key_columns, parameter_columns, value_columns) -> LookupTable:
        return LookupTable(data, self._pop_view_builder, key_columns, parameter_columns,
                           value_columns, self._interpolation_order, self.clock, self._extrapolate)

    def __repr__(self):
        return "LookupTableManager()"


class LookupTableInterface:

    def __init__(self, manager):
        self._lookup_table_manager = manager

    def build_table(self, data, key_columns=('sex',), parameter_columns=(['age', 'age_group_start', 'age_group_end'],
                                                                         ['year', 'year_start', 'year_end']),
                    value_columns=None) -> LookupTable:
        """Construct a LookupTable from input data.

        If data is a ``pandas.DataFrame``, an interpolation function of the specified
        order will be calculated for each permutation of the set of key_columns.
        The columns in parameter_columns will be used as parameters for the interpolation
        functions which will estimate all remaining columns in the table.

        If data is a number, time, list, or tuple, a scalar table will be constructed with
        the values in data as the values in each column of the table, named according to
        value_columns.


        Parameters
        ----------
        data        : The source data which will be used to build the resulting LookupTable.
        key_columns : [str]
                      Columns used to select between interpolation functions. These
                      should be the non-continuous variables in the data. For
                      example 'sex' in data about a population.
        parameter_columns : [str]
                      The columns which contain the parameters to the interpolation functions.
                      These should be the continuous variables. For example 'age'
                      in data about a population.
        value_columns : [str]
                      The data columns that will be in the resulting LookupTable. Columns to be
                      interpolated over if interpolation or the names of the columns in the scalar
                      table.
        Returns
        -------
        LookupTable
        """
        return self._lookup_table_manager.build_table(data, key_columns, parameter_columns, value_columns)



import os

import pytest

from vivarium.framework.configuration import build_simulation_configuration
from vivarium.framework.components.manager import (ComponentManager, ComponentConfigError, OrderedComponentSet,
                                                   setup_components, apply_component_default_configuration)

from .mocks import MockComponentA, MockComponentB, NamelessComponent


@pytest.fixture
def apply_default_config_mock(mocker):
    return mocker.patch('vivarium.framework.components.manager.apply_component_default_configuration')


def test_apply_component_default_configuration():

    class UnladenSwallow:

        configuration_defaults = {
            'unladen_swallow': {
                'airspeed_velocity': 11,
            }
        }

    us = UnladenSwallow()
    config = build_simulation_configuration()
    assert 'unladen_swallow' not in config
    apply_component_default_configuration(config, us)
    assert config.unladen_swallow.metadata('airspeed_velocity') == [
        {'layer': 'component_configs', 'value': 11,
         'source': os.path.realpath(__file__), 'default': False}
    ]

    us = UnladenSwallow()
    us.__module__ = '__main__'
    config = build_simulation_configuration()
    assert 'unladen_swallow' not in config
    apply_component_default_configuration(config, us)
    assert config.unladen_swallow.metadata('airspeed_velocity') == [
        {'layer': 'component_configs', 'value': 11, 'source': '__main__', 'default': False}
    ]


def test_setup_components(mocker, apply_default_config_mock):
    config = build_simulation_configuration()
    builder = mocker.Mock()

    components = OrderedComponentSet(MockComponentA('Eric'), MockComponentB('half', 'a', 'bee'))
    finished = setup_components(builder, components, config)
    mock_a, mock_b = finished

    apply_default_config_mock.assert_not_called()

    assert mock_a.builder_used_for_setup is None  # class has no setup method
    assert mock_b.builder_used_for_setup is builder
    assert mock_b.args == ('half', 'a', 'bee')

    class TestBee:
        configuration_defaults = {
            'test_bee': {
                'name': 'wasp',
                'location': 'next to my bed',
            }
        }

        def __init__(self):
            self.name = 'TestBee'

    test_bee = TestBee()
    components = OrderedComponentSet(MockComponentA('Eric'), MockComponentB('half', 'a', 'bee'), test_bee)
    apply_default_config_mock.reset_mock()
    setup_components(builder, components, config)

    apply_default_config_mock.assert_called_once()


@pytest.mark.parametrize("components", (
        (MockComponentA('Eric'), MockComponentB('half', 'a', 'bee')),
        (MockComponentA('Eric'),)
))
def test_ComponentManager__add_components(components):
    manager = ComponentManager()

    for list_type in ['_managers', '_components']:
        manager._add_components(getattr(manager, list_type), components)
        assert getattr(manager, list_type) == OrderedComponentSet(*components)


@pytest.mark.parametrize("components", (
        (MockComponentA(), MockComponentA()),
        (MockComponentA(), MockComponentA(), MockComponentB('foo', 'bar')),
))
def test_ComponentManager__add_components_duplicated(components):
    manager = ComponentManager()

    for list_type in ['_managers', '_components']:
        with pytest.raises(ComponentConfigError, match='duplicate name'):
            manager._add_components(getattr(manager, list_type), components)


@pytest.mark.parametrize("components", (
        (None,),
        (NamelessComponent(),),
        (NamelessComponent(), MockComponentA())
))
def test_ComponentManager__add_components_unnamed(components):
    manager = ComponentManager()

    for list_type in ['_managers', '_components']:
        with pytest.raises(ComponentConfigError, match='no name'):
            manager._add_components(getattr(manager, list_type), components)


def test_ComponentManager__setup_components(mocker):
    config = build_simulation_configuration()
    manager = ComponentManager()
    builder = mocker.Mock()
    builder.components = manager

    # manager._components = []
    manager.add_components([MockComponentA('Eric'), MockComponentB('half', 'a', 'bee')])
    manager.setup_components(builder, config)

    mock_a, mock_b, mock_b_child1, mock_b_child2, mock_b_child3 = manager._components

    assert mock_a.builder_used_for_setup is None  # class has no setup method
    assert mock_b.builder_used_for_setup is builder
    assert mock_b_child1.args == ('half',)
    assert mock_b_child1.builder_used_for_setup is builder
    assert mock_b_child2.args == ('a',)
    assert mock_b_child2.builder_used_for_setup is builder
    assert mock_b_child3.args == ('bee',)
    assert mock_b_child3.builder_used_for_setup is builder


class DummyMachine:
    configuration_defaults = {
        'factory': {
            'location': 'Building A'
        },
        'dummy_machine': {
            'id': 1,
        }
    }

    def __init__(self):
        self.name = 'dummy_machine'


class DummyMechanic:
    configuration_defaults = {
        'work_level': {
            'capacity': 2,  #per day
            'success_rate': .96
        },
        'dummy_machine': {
            'id': 123,
        }
    }

    def __init__(self):
        self.name = 'dummy_mechanic'


def test_default_configuration_set_by_one_component(mocker):

    machine = DummyMachine()
    mechanic = DummyMechanic()

    manager = ComponentManager()
    builder = mocker.Mock()
    builder.components = manager
    config = build_simulation_configuration()

    for key in ['factory', 'dummy_machine', 'work_level']:
        assert key not in config

    manager.add_components([machine, mechanic])

    with pytest.raises(ComponentConfigError):
        manager.setup_components(builder, config)


def test_default_configuration_set_by_one_component_as_0(mocker):

    DummyMachine.configuration_defaults['dummy_machine']['id'] = 0
    machine = DummyMachine()
    mechanic = DummyMechanic()

    manager = ComponentManager()
    builder = mocker.Mock()
    builder.components = manager
    config = build_simulation_configuration()

    manager.add_components([machine, mechanic])

    with pytest.raises(ComponentConfigError):
        manager.setup_components(builder, config)

    # case with empty string as default config value
    DummyMachine.configuration_defaults['dummy_machine']['id'] = ''
    machine = DummyMachine()
    manager = ComponentManager()
    builder = mocker.Mock()
    builder.components = manager
    config = build_simulation_configuration()
    manager.add_components([machine, mechanic])

    with pytest.raises(ComponentConfigError):
        manager.setup_components(builder, config)


def test_default_configuration_set_by_one_component_different_tree_depths(mocker):

    DummyMachine.configuration_defaults['dummy_machine']['id'] = {'number': 1, 'type': 'machine'}
    machine = DummyMachine()
    mechanic = DummyMechanic()

    manager = ComponentManager()
    builder = mocker.Mock()
    builder.components = manager
    config = build_simulation_configuration()

    manager.add_components([machine, mechanic])

    with pytest.raises(ComponentConfigError):
        manager.setup_components(builder, config)


def test_ComponentSet_add():
    component_list = OrderedComponentSet()

    component_0 = MockComponentA(name='component_0')
    component_list.add(component_0)

    component_1 = MockComponentA(name='component_1')
    component_list.add(component_1)

    # duplicates by name
    with pytest.raises(ComponentConfigError, match='duplicate name'):
        component_list.add(component_0)

    # no name
    with pytest.raises(ComponentConfigError, match='no name'):
        component_list.add(NamelessComponent())


def test_ComponentSet_update():
    component_list = OrderedComponentSet()

    components = [MockComponentA(name='component_0'), MockComponentA('component_1')]

    component_list.update(components)

    with pytest.raises(ComponentConfigError, match='duplicate name'):
        component_list.update(components)
    with pytest.raises(ComponentConfigError, match='no name'):
        component_list.update([NamelessComponent()])


def test_ComponentSet_initialization():
    component_1 = MockComponentA()
    component_2 = MockComponentB()

    component_list = OrderedComponentSet(component_1, component_2)
    assert component_list.components == [component_1, component_2]


def test_ComponentSet_pop():
    component = MockComponentA()
    component_list = OrderedComponentSet(component)

    c = component_list.pop()
    assert c == component

    with pytest.raises(IndexError):
        component_list.pop()


def test_ComponentSet_contains():
    component_list = OrderedComponentSet()

    assert not bool(component_list)
    assert len(component_list) == 0

    component_1 = MockComponentA()
    component_2 = MockComponentB()
    component_3 = MockComponentA(name='absent')
    component_list = OrderedComponentSet(component_1, component_2)

    assert component_1 in component_list
    assert component_3 not in component_list

    with pytest.raises(ComponentConfigError, match='no name'):
        throwaway = 10 in component_list


def test_ComponentSet_eq():
    component_1 = MockComponentA()
    component_2 = MockComponentB()
    component_list = OrderedComponentSet(component_1, component_2)

    assert component_list == component_list
    assert component_list != 10

    second_list = OrderedComponentSet(component_1)
    assert component_list != second_list


def test_ComponentSet_bool_len():
    component_list = OrderedComponentSet()

    assert not bool(component_list)
    assert len(component_list) == 0

    component_1 = MockComponentA()
    component_2 = MockComponentB()
    component_list = OrderedComponentSet(component_1, component_2)

    assert bool(component_list)
    assert len(component_list) == 2



"""
==========================
Vivarium Interactive Tools
==========================

This module provides an interface for interactive simulation usage. The main
part is the :class:`InteractiveContext`, a sub-class of the main simulation
object in ``vivarium`` that has been extended to include convenience
methods for running and exploring the simulation in an interactive setting.

Rather than initializing the :class:`InteractiveContext` directly, users should
acquire one using one of the four simulation creation
:ref:`functions <simulation_creation>` provided
in this module.

See the associated tutorials for :ref:`running <interactive_tutorial>` and
:ref:`exploring <exploration_tutorial>` for more information.

"""
from math import ceil
from typing import Mapping, List, Callable, Dict, Any
import warnings

import pandas as pd

from vivarium.framework.configuration import build_simulation_configuration, build_model_specification, ConfigTree
from vivarium.framework.engine import SimulationContext
from vivarium.framework.plugins import PluginManager
from vivarium.framework.time import Timedelta, Time
from vivarium.framework.values import Pipeline

from .utilities import run_from_ipython, log_progress, raise_if_not_setup


class InteractiveContext(SimulationContext):
    """A simulation context with helper methods for running simulations
    interactively.

    This class should not be instantiated directly. It should be created with a
    call to one of the helper methods provided in this module.
    """

    def __init__(self, configuration: ConfigTree, components: List, plugin_manager: PluginManager = None):
        super().__init__(configuration, components, plugin_manager)
        self._initial_population = None

    def setup(self):
        """Setup the simulation and initialize its population."""
        super().setup()
        self._start_time = self.clock.time
        self.initialize_simulants()

    def initialize_simulants(self):
        """Initialize this simulation's population.

        This method should be called by the framework, not by the user.
        """
        super().initialize_simulants()
        self._initial_population = self.population.get_population(True)

    @raise_if_not_setup(system_type='run')
    def reset(self):
        """Reset the simulation to its initial state."""
        warnings.warn("This reset method is very crude.  It should work for "
                      "many simple simulations, but we make no guarantees. In "
                      "particular, if you have components that manage their "
                      "own state in any way, this might not work.")
        self.population._population = self._initial_population
        self.clock._time = self._start_time

    @raise_if_not_setup(system_type='run')
    def run(self, with_logging: bool = True) -> int:
        """Run the simulation for the duration specified in the configuration.

        Parameters
        ----------
        with_logging
            Whether or not to log the simulation steps. Only works in an ipython
            environment.

        Returns
        -------
            The number of steps the simulation took.
        """
        return self.run_until(self.clock.stop_time, with_logging=with_logging)

    @raise_if_not_setup(system_type='run')
    def run_for(self, duration: Timedelta, with_logging: bool = True) -> int:
        """Run the simulation for the given time duration.

        Parameters
        ----------
        duration
            The length of time to run the simulation for. Should be the same
            type as the simulation clock's step size (usually a pandas
            Timedelta).
        with_logging
            Whether or not to log the simulation steps. Only works in an ipython
            environment.

        Returns
        -------
            The number of steps the simulation took.
        """
        return self.run_until(self.clock.time + duration, with_logging=with_logging)

    @raise_if_not_setup(system_type='run')
    def run_until(self, end_time: Time, with_logging: bool = True) -> int:
        """Run the simulation until the provided end time.

        Parameters
        ----------
        end_time
            The time to run the simulation until. The simulation will run until
            its clock is greater than or equal to the provided end time.
        with_logging
            Whether or not to log the simulation steps. Only works in an ipython
            environment.

        Returns
        -------
            The number of steps the simulation took.
        """
        if not isinstance(end_time, type(self.clock.time)):
            raise ValueError(f"Provided time must be an instance of {type(self.clock.time)}")

        iterations = int(ceil((end_time - self.clock.time)/self.clock.step_size))
        self.take_steps(number_of_steps=iterations, with_logging=with_logging)
        assert self.clock.time - self.clock.step_size < end_time <= self.clock.time
        return iterations

    @raise_if_not_setup(system_type='run')
    def step(self, step_size: Timedelta = None):
        """Advance the simulation one step.

        Parameters
        ----------
        step_size
            An optional size of step to take. Must be the same type as the
            simulation clock's step size (usually a pandas.Timedelta).
        """
        old_step_size = self.clock.step_size
        if step_size is not None:
            if not isinstance(step_size, type(self.clock.step_size)):
                raise ValueError(f"Provided time must be an instance of {type(self.clock.step_size)}")
            self.clock._step_size = step_size
        super().step()
        self.clock._step_size = old_step_size

    @raise_if_not_setup(system_type='run')
    def take_steps(self, number_of_steps: int = 1, step_size: Timedelta = None, with_logging: bool = True):
        """Run the simulation for the given number of steps.

        Parameters
        ----------
        number_of_steps
            The number of steps to take.
        step_size
            An optional size of step to take. Must be the same type as the
            simulation clock's step size (usually a pandas.Timedelta).
        with_logging
            Whether or not to log the simulation steps. Only works in an ipython
            environment.
        """
        if not isinstance(number_of_steps, int):
            raise ValueError('Number of steps must be an integer.')

        if run_from_ipython() and with_logging:
            for _ in log_progress(range(number_of_steps), name='Step'):
                self.step(step_size)
        else:
            for _ in range(number_of_steps):
                self.step(step_size)

    @raise_if_not_setup(system_type='population')
    def get_population(self, untracked: bool = False) -> pd.DataFrame:
        """Get a copy of the population state table.

        Parameters
        ----------
        untracked
            Whether or not to return simulants who are no longer being tracked
            by the simulation.
        """
        return self.population.get_population(untracked)

    @raise_if_not_setup(system_type='value')
    def list_values(self) -> List[str]:
        """List the names of all pipelines in the simulation."""
        return list(self.values.keys())

    @raise_if_not_setup(system_type='value')
    def get_value(self, value_pipeline_name: str) -> Pipeline:
        """Get the value pipeline associated with the given name."""
        return self.values.get_value(value_pipeline_name)

    @raise_if_not_setup(system_type='event')
    def list_events(self) -> List[str]:
        """List all event types registered with the simulation."""
        return self.events.list_events()

    @raise_if_not_setup(system_type='event')
    def get_listeners(self, event_type: str) -> List[Callable]:
        """Get all listeners of a particular type of event.

        Available event types can be found by calling
        :func:`InteractiveContext.list_events`.

        Parameters
        ----------
        event_type
            The type of event to grab the listeners for.

        """
        if event_type not in self.events:
            raise ValueError(f'No event {event_type} in system.')
        return self.events.get_listeners(event_type)

    @raise_if_not_setup(system_type='event')
    def get_emitter(self, event_type: str) -> Callable:
        """Get the callable that emits the given type of events.

        Available event types can be found by calling
        :func:`InteractiveContext.list_events`.

        Parameters
        ----------
        event_type
            The type of event to grab the listeners for.

        """
        if event_type not in self.events:
            raise ValueError(f'No event {event_type} in system.')
        return self.events.get_emitter(event_type)

    @raise_if_not_setup(system_type='component')
    def list_components(self) -> Dict[str, Any]:
        """Get a mapping of component names to components currently in the simulation.

        Returns
        -------
            A dictionary mapping component names to components.

        """
        return  self.component_manager.list_components()

    @raise_if_not_setup(system_type='component')
    def get_component(self, name: str) -> Any:
        """Get the component in the simulation that has ``name``, if present.
        Names are guaranteed to be unique.

        Parameters
        ----------
        name
            A component name.
        Returns
        -------
            A component that has the name ``name`` else None.

        """
        return self.component_manager.get_component(name)


def initialize_simulation(components: List, input_config: Mapping = None,
                          plugin_config: Mapping = None) -> InteractiveContext:
    """Construct a simulation from a list of components, component
    configuration, and a plugin configuration.

    The simulation context returned by this method still needs to be setup by
    calling its setup method. It is mostly useful for testing and debugging.

    Parameters
    ----------
    components
        A list of initialized simulation components. Corresponds to the
        components block of a model specification.
    input_config
        A nested dictionary with any additional simulation configuration
        information needed. Corresponds to the configuration block of a model
        specification.
    plugin_config
        A dictionary containing a description of any simulation plugins to
        include in the simulation. If you're using this argument, you're either
        deep in the process of simulation development or the maintainers have
        done something wrong. Corresponds to the plugins block of a model
        specification.

    Returns
    -------
        An initialized (but not set up) simulation context.
    """
    config = build_simulation_configuration()
    config.update(input_config)
    plugin_manager = PluginManager(plugin_config)

    return InteractiveContext(config, components, plugin_manager)


def setup_simulation(components: List, input_config: Mapping = None,
                     plugin_config: Mapping = None) -> InteractiveContext:
    """Construct a simulation from a list of components and call its setup
    method.

    Parameters
    ----------
    components
        A list of initialized simulation components. Corresponds to the
        components block of a model specification.
    input_config
        A nested dictionary with any additional simulation configuration
        information needed. Corresponds to the configuration block of a model
        specification.
    plugin_config
        A dictionary containing a description of any simulation plugins to
        include in the simulation. If you're using this argument, you're either
        deep in the process of simulation development or the maintainers have
        done something wrong. Corresponds to the plugins block of a model
        specification.

    Returns
    -------
        A simulation context that is setup and ready to run.
    """
    simulation = initialize_simulation(components, input_config, plugin_config)
    simulation.setup()

    return simulation


def initialize_simulation_from_model_specification(model_specification_file: str) -> InteractiveContext:
    """Construct a simulation from a model specification file.

    The simulation context returned by this method still needs to be setup by
    calling its setup method. It is mostly useful for testing and debugging.

    Parameters
    ----------
    model_specification_file
        The path to a model specification file.

    Returns
    -------
        An initialized (but not set up) simulation context.
    """
    model_specification = build_model_specification(model_specification_file)

    plugin_config = model_specification.plugins
    component_config = model_specification.components
    simulation_config = model_specification.configuration

    plugin_manager = PluginManager(plugin_config)
    component_config_parser = plugin_manager.get_plugin('component_configuration_parser')
    components = component_config_parser.get_components(component_config)

    return InteractiveContext(simulation_config, components, plugin_manager)


def setup_simulation_from_model_specification(model_specification_file: str) -> InteractiveContext:
    """Construct a simulation from a model specification file and call its setup
    method.

    Parameters
    ----------
    model_specification_file
        The path to a model specification file.

    Returns
    -------
        A simulation context that is setup and ready to run.
    """
    simulation = initialize_simulation_from_model_specification(model_specification_file)
    simulation.setup()

    return simulation



import numpy
numpy.seterr(all='raise')

from vivarium.config_tree import ConfigTree
from vivarium.interface import (build_simulation_configuration, setup_simulation,
                                setup_simulation_from_model_specification, initialize_simulation,
                                initialize_simulation_from_model_specification)
from vivarium.__about__ import (__author__, __copyright__, __email__, __license__,
                                __summary__, __title__, __uri__, __version__, )



"""
============================
The Component Manager System
============================
The ``vivarium`` component manager system is responsible for maintaining a
reference to all of the managers and components in a simulation, providing an
interface for adding additional components or managers, and applying default
configurations and initiating the ``setup`` stage of the lifecycle. This module
provides the default implementation and interface.

The :class:`ComponentManager` is the first plugin loaded by the
:class:`SimulationContext <vivarium.framework.engine.SimulationContext>`
and managers and components are given to it by the context. It is called on to
setup everything it holds when the context itself is setup.

"""
import inspect
from typing import Union, List, Tuple, Iterator, Dict, Any, Type

from vivarium.config_tree import ConfigTree
from vivarium.exceptions import VivariumError


class ComponentConfigError(VivariumError):
    """Error while interpreting configuration file or initializing components"""
    pass


class OrderedComponentSet:
    """A container for Vivarium components. It preserves ordering, enforces
    uniqueness by name, and provides a subset of set-like semantics."""

    def __init__(self, *args):
        self.components = []
        if args:
            self.update(args)

    def add(self, component: Any):
        if component in self:
            raise ComponentConfigError(f"Attempting to add a component with duplicate name: {component}")
        self.components.append(component)

    def update(self, components: Union[List[Any], Tuple[Any]]):
        for c in components:
            self.add(c)

    def __contains__(self, component: Any) -> bool:
        if not hasattr(component, "name"):
            raise ComponentConfigError(f"Component {component} has no name attribute")
        return component.name in [c.name for c in self.components]

    def __iter__(self) -> Iterator:
        return iter(self.components)

    def __len__(self) -> int:
        return len(self.components)

    def __bool__(self) -> bool:
        return bool(self.components)

    def __eq__(self, other) -> bool:
        try:
            return type(self) is type(other) and [c.name for c in self.components] == [c.name for c in other.components]
        except TypeError:
            return False

    def pop(self) -> Any:
        component = self.components.pop(0)
        return component

    def __repr__(self):
        return "OrderedComponentSet()"


class ComponentManager:
    """Maintains references to all components and managers in a ``vivarium``
    simulation, applies their default configuration and initiates their
    ``setup`` life-cycle stage.

    The component manager maintains a separate list of managers and components
    and provides methods for adding to these lists and getting members that
    correspond to a specific type.  It also initiates the ``setup`` lifecycle
    phase for all components and managers it controls. This is done first for
    managers and then components, and involves applying default configurations
    and calling the object's ``setup`` method.

    """

    def __init__(self):
        self._managers = OrderedComponentSet()
        self._components = OrderedComponentSet()

    @property
    def name(self):
        return "component_manager"

    def add_managers(self, managers: Union[List[Any], Tuple[Any]]):
        """Registers new managers with the component manager. Managers are
        configured and setup before components.

        Parameters
        ----------
        managers:
          Instantiated managers to register.
        """
        self._add_components(self._managers, managers)

    def add_components(self, components: Union[List[Any], Tuple[Any]]):
        """Register new components with the component manager. Components are
        configured and setup after managers.

        Parameters
        ----------
        components:
          Instantiated components to register.
        """
        self._add_components(self._components, components)

    def _add_components(self, component_list: OrderedComponentSet, components: Union[List[Any], Tuple[Any]]):
        for component in components:
            if isinstance(component, list) or isinstance(component, tuple):
                self._add_components(component_list, component)
            else:
                component_list.add(component)

    def get_components_by_type(self, component_type: Union[type, Tuple[type]]) -> List[Any]:
        """Get all components currently held by the component manager that are an
        instance of ``component_type``.

        Parameters
        ----------
        component_type
            A component type.
        Returns
        -------
            A list of components of type ``component_type``.
        """
        return [c for c in self._components if isinstance(c, component_type)]

    def get_component(self, name: str) -> Any:
        """Get the component that has ``name`` if presently held by the component
        manager. Names are guaranteed to be unique.

        Parameters
        ----------
        name
            A component name.
        Returns
        -------
            A component that has name ``name``.
        Raises
        ------
        ValueError
            No component exists in the component manager with ``name``.
        """
        for c in self._components:
            if c.name == name:
                return c
        raise ValueError(f"No component found with name {name}")

    def list_components(self) -> Dict[str, Any]:
        """Get a mapping of component names to components held by the manager.

        Returns
        -------
            A mapping of component names to components.
        """
        return {c.name: c for c in self._components}

    def setup_components(self, builder, configuration: ConfigTree):
        """Separately configure and set up the managers and components held by
        the component manager, in that order.

        The setup process involves applying default configurations and then
        calling the manager or component's setup method. This can result in new
        components as a side effect of setup because components themselves have
        access to this interface through the builder in their setup method.

        Parameters
        ----------
        builder:
            Interface to several simulation tools.
        configuration:
            Simulation configuration parameters.
        """
        self._managers = setup_components(builder, self._managers, configuration)
        self._components = setup_components(builder, self._components, configuration)

    def __repr__(self):
        return f"ComponentManager()"


class ComponentInterface:
    """The builder interface for the component manager system. This class
    defines component manager methods a ``vivarium`` component can access from
    the builder. It provides methods for querying and adding components to the
    :class:`ComponentManager`.

    """

    def __init__(self, component_manager: ComponentManager):
        self._component_manager = component_manager

    def add_components(self, components: Union[List[Any], Tuple[Any]]):
        """Register new components with the component manager. Components are
        configured  and setup after managers.

        Parameters
        ----------
        components:
          Instantiated components to register.

        """
        self._component_manager.add_components(components)

    def get_component(self, name: str) -> Any:
        """Get the component that has ``name`` if presently held by the component
        manager. Names are guaranteed to be unique.

        Parameters
        ----------
        name
            A component name.
        Returns
        -------
            A component that has name ``name``.

        """
        return self._component_manager.get_component(name)

    def get_components_by_type(self, component_type: Type) -> List[Any]:
        """Get all components that are an instance of ``component_type``
        currently held by the component manager.

        Parameters
        ----------
        component_type
            A component type to retrieve, compared against internal components
            using isinstance().
        Returns
        -------
            A list of components of type ``component_type``.

        """
        return self._component_manager.get_components_by_type(component_type)

    def list_components(self) -> Dict[str, Any]:
        """Get a mapping of component names to components held by the manager.

        Returns
        -------
            A dictionary mapping component names to components.

        """
        return self._component_manager.list_components()


def setup_components(builder, component_list: OrderedComponentSet, configuration: ConfigTree) -> OrderedComponentSet:
    """Configure and set up a list of components or managers.

    This function first loops over ``component_list`` and applies configuration
    defaults if present, modifying the ``configuration`` object in the process.
    Then, the ``component_list`` is looped over until empty, with components
    being popped and setup. Because of the side effect described below, new
    components can be added to the list. This necessitates a while loop and a
    check against unconfigured components.

    Importantly, this function is called by the :class:`ComponentManager` and it
    is passed a reference to one of the lists that manager holds. This list is
    also accessible through the :class:`ComponentInterface` on the builder, so a
    component's setup method can mutate the list.

    Parameters
    ----------
    builder
        Interface to several simulation tools.
    component_list
        A list of vivarium components.
    configuration
        A vivarium configuration object.

    Returns
    -------
        A list of components which have been configured and setup.
    """
    configured = []
    for c in component_list:  # apply top-level configurations first
        if hasattr(c, "configuration_defaults") and c not in configured:
            apply_component_default_configuration(configuration, c)
            configured.append(c)

    setup = []
    while component_list:  # mutated at runtime by calls to setup
        c = component_list.pop()
        if c is None:
            raise ComponentConfigError('None in component list. This likely '
                                       'indicates a bug in a factory function')

        if hasattr(c, "configuration_defaults") and c not in configured:
            apply_component_default_configuration(configuration, c)
            configured.append(c)

        if c not in setup:
            if hasattr(c, "setup"):
                c.setup(builder)
            setup.append(c)

    return OrderedComponentSet(*setup)


def apply_component_default_configuration(configuration: ConfigTree, component: Any):
    """Check if a default configuration attached to a component is duplicated
    in the simulation configuration, which will raise. If it is not, apply the
    configuration.

    Parameters
    ----------
    configuration
        A vivarium configuration object.
    component
        A vivarium component.

    """

    # This reapplies configuration from some components but it is idempotent so there's no effect.
    if component.__module__ == '__main__':
        # This is defined directly in a script or notebook so there's no file to attribute it to
        source = '__main__'
    else:
        source = inspect.getfile(component.__class__)
    check_duplicated_default_configuration(component.configuration_defaults, configuration, source)
    configuration.update(component.configuration_defaults, layer='component_configs', source=source)


def check_duplicated_default_configuration(component: Any, config: ConfigTree, source: str):
    """Check that the keys present in a component's default configuration
    ``component`` are not already present in the global configtree ``config``.

    Parameters
    ----------
    component
        A vivarium component.
    config
        A vivarium configuration object.
    source
        The file containing the code that describes the component. Only used
        to generate a cogent error message.

    Raises
    -------
    ComponentConfigError
        A component's default configuration is already present in the config tree.

    """
    overlapped = set(component.keys()).intersection(config.keys())
    if not overlapped:
        pass

    while overlapped:
        key = overlapped.pop()

        try:
            sub_config = config.get_from_layer(key, layer='component_configs')
            sub_component = component[key]

            if isinstance(sub_component, dict) and isinstance(sub_config, ConfigTree):
                check_duplicated_default_configuration(sub_component, sub_config, source)
            elif isinstance(sub_component, dict) or isinstance(sub_config, ConfigTree):
                raise ComponentConfigError(f'These two sources have different structure of configurations for {component}.'
                                           f' Check {source} and {sub_config}')
            else:
                raise ComponentConfigError(f'Check these two {source} and {config._children[key].get_value_with_source()}'
                                           f'Both try to set the default configurations for {component}/{key}')

        except KeyError:
            pass
